/*
 Copyright (c) Jupyter Development Team.
 Distributed under the terms of the Modified BSD License.
*/
@import '@lumino/default-theme/style/index.css';
@import './content.css';

body {
  display: flex;
  flex-direction: column;
  position: absolute;
  top: 0;
  left: 0;
  right: 0;
  bottom: 0;
  margin: 0;
  padding: 0;
  overflow: hidden;
}

#main {
  flex: 1 1 auto;
}

#accordion {
  flex: 1 1 auto;
  padding: 4px;
}
