/*
 * Copyright (c) Jupyter Development Team.
 * Distributed under the terms of the Modified BSD License.
 */

/*-----------------------------------------------------------------------------
| Copyright (c) Jupyter Development Team.
| Copyright (c) 2014-2017, PhosphorJS Contributors
|
| Distributed under the terms of the BSD 3-Clause License.
|
| The full license is in the file LICENSE, distributed with this software.
|----------------------------------------------------------------------------*/

.lm-TabBar {
  min-height: 24px;
  max-height: 24px;
}

.lm-TabBar-content {
  min-width: 0;
  min-height: 0;
  align-items: flex-end;
  border-bottom: 1px solid #c0c0c0;
}

.lm-TabBar-tab {
  padding: 0px 10px;
  background: #e5e5e5;
  border: 1px solid #c0c0c0;
  border-bottom: none;
  font:
    12px Helvetica,
    Arial,
    sans-serif;
  flex: 0 1 125px;
  min-height: 20px;
  max-height: 20px;
  min-width: 35px;
  margin-left: -1px;
  line-height: 20px;
}

.lm-TabBar-tabLabel .lm-TabBar-tabInput {
  padding: 0px;
  border: 0px;
  font:
    12px Helvetica,
    Arial,
    sans-serif;
}

.lm-TabBar-tab.lm-mod-current {
  background: white;
}

.lm-TabBar-tab:hover:not(.lm-mod-current) {
  background: #f0f0f0;
}

.lm-TabBar-tab:first-child {
  margin-left: 0;
}

.lm-TabBar-tab.lm-mod-current {
  min-height: 23px;
  max-height: 23px;
  transform: translateY(1px);
}

.lm-TabBar-tabIcon,
.lm-TabBar-tabLabel,
.lm-TabBar-tabCloseIcon {
  display: inline-block;
}

.lm-TabBar-tab.lm-mod-closable > .lm-TabBar-tabCloseIcon {
  margin-left: 4px;
}

.lm-TabBar .lm-TabBar-addButton {
  padding: 0px 6px;
  border-bottom: 1px solid #c0c0c0;
}

.lm-TabBar-tab.lm-mod-closable > .lm-TabBar-tabCloseIcon:before {
  content: '\f00d';
  font-family: FontAwesome;
}

.lm-TabBar .lm-TabBar-addButton:before {
  content: '\f067';
  font-family: FontAwesome;
}

.lm-TabBar-tab.lm-mod-drag-image {
  min-height: 23px;
  max-height: 23px;
  min-width: 125px;
  border: none;
  box-shadow: 1px 1px 2px rgba(0, 0, 0, 0.3);
}
