/*
 Copyright (c) Jupyter Development Team.
 Distributed under the terms of the Modified BSD License.
*/

.content {
  min-width: 50px;
  min-height: 50px;
  display: flex;
  flex-direction: column;
  padding: 8px;
  border: 1px solid #c0c0c0;
  background: white;
  box-shadow: 1px 1px 2px rgba(0, 0, 0, 0.2);
}

.lm-AccordionPanel[data-orientation='vertical'] .content {
  border-top: none;
}

.lm-AccordionPanel[data-orientation='horizontal'] .content {
  border-left: none;
}

.content > div {
  flex: 1 1 auto;
  border: 1px solid #505050;
  overflow: auto;
}

.content input {
  margin: 8px;
}

.red > div {
  background: #e74c3c;
}

.yellow > div {
  background: #f1c40f;
}

.green > div {
  background: #27ae60;
}

.blue > div {
  background: #3498db;
}
