// Copyright (c) Jupyter Development Team.
// Distributed under the terms of the Modified BSD License.
/*-----------------------------------------------------------------------------
| Copyright (c) 2014-2017, PhosphorJS Contributors
|
| Distributed under the terms of the BSD 3-Clause License.
|
| The full license is in the file LICENSE, distributed with this software.
|----------------------------------------------------------------------------*/
import '@lumino/widgets/style/index.css';

import './accordionlayout.spec';
import './accordionpanel.spec';
import './boxengine.spec';
import './boxlayout.spec';
import './boxpanel.spec';
import './commandpalette.spec';
import './contextmenu.spec';
import './docklayout.spec';
import './dockpanel.spec';
import './focustracker.spec';
import './layout.spec';
import './menu.spec';
import './menubar.spec';
import './panel.spec';
import './panellayout.spec';
import './splitlayout.spec';
import './splitpanel.spec';
import './stackedlayout.spec';
import './stackedpanel.spec';
import './tabbar.spec';
import './tabpanel.spec';
import './title.spec';
import './widget.spec';
